import re

from selenium.webdriver.common.by import By
from selenium.webdriver.support.wait import TimeoutException

from .util import Condition, Result

from .tables import Table

info_re = re.compile(ur"^Showing (\d+) to (\d+) of (\d+) entries")

_REDRAW_SETUP_FUNCTION = """
function redrawSetup(cssid, done) {
    window.__selenium_test_redrawn = false;
    var processing = document.getElementById(cssid + "_processing");
    var table = document.getElementById(cssid);
    function check() {
        // We do this to make sure that the table is not currently
        // refreshing when we put in our event handler.
        if (processing.style.display !== "none") {
            setTimeout(check, 100);
            return;
        }
        jQuery(table).one("draw.dt", function () {
            window.__selenium_test_redrawn = true;
        });
        done();
    }
    check();
};
"""

_REDRAW_SETUP_SNIPPET = _REDRAW_SETUP_FUNCTION + """
redrawSetup(arguments[0], arguments[1]);
"""

_REDRAW_CHECK_SNIPPET = """
var cb = arguments[0];
function test() {
    if (window.__selenium_test_redrawn) {
        cb();
        return;
    }
    setTimeout(test, 100);
}
test();
"""

class Datatable(Table):

    def __init__(self, *args, **kwargs):
        super(Datatable, self).__init__(*args, **kwargs)
        self.field_selectors.append("#" + self.cssid + "_filter")

    def wait_for_initialized(self):
        try:
            self.util.driver.execute_async_script("""
            var cssid = arguments[0];
            var done = arguments[1];
            var $table = jQuery(document.getElementById(cssid));

            function check() {
              // When Datatables may be loaded by a module loader,
              // it is possible that Datatables has not been loaded
              // yet. Poll.
              if (!jQuery.fn.dataTable) {
                setTimeout(check, 100);
                return;
              }

              var settings = jQuery.fn.dataTable.isDataTable($table) &&
                $table.DataTable().settings()[0];
              if (settings && settings._bInitComplete) {
                // Already initialized.
                done();
                return;
              }
              // Not initialized: wait for the init event.
              $table.one("init.dt", function () {
                done();
              });
            }

            check();
            """, self.cssid)
            # If we did not timeout, then the table is initialized
            return True
        except TimeoutException:
            # If we did timeout, initialization did not happen.
            return False

    def setup_redraw_check(self):
        self.util.driver.execute_async_script(_REDRAW_SETUP_SNIPPET,
                                              self.cssid)

    def wait_for_redraw(self):
        self.util.driver.execute_async_script(_REDRAW_CHECK_SNIPPET)

    def wait_for_results(self, expected_total):
        def check(driver):
            text = driver.find_element_by_id(self.cssid + "_info").text
            match = info_re.match(text)
            if not match:
                return False

            total = int(match.group(3))
            return Result(total == expected_total, total)

        result = Condition(self.util, check).wait()

        return result.payload

    def get_result(self, number):
        return self.util.find_elements((By.CSS_SELECTOR,
                                        "#" + self.cssid + ">tbody>tr"))[number]



"""
"""

class NormalizedCapabilities(dict):

    def __init__(self, caps):
        """
        Once upon a time it was possible to just query the capabilities
        returned from a ``WebDriver`` object without having to worry
        about which browser was being driven. The dictionary returned
        would have the same fields across browsers.

        The driver for Edge and geckodriver changed all this. For
        instance, they store the browser version under
        ``browserVersion`` instead of ``version`` everywhere else. And
        they store the platform name under ``platformName`` instead of
        ``platform``.

        The variations are very problematic for code that needs to
        query the capabilities but should work uniformly across
        browsers.

        Instances of this class present a normalized view of the
        capabilities. Instances should be treated as read-only. They
        contains the same fields as the capabilities passed in the
        constructor, with the following differences:

        * ``version`` is renamed ``browserVersion``

        * ``platform`` is renamed ``platformName``

        We normalize to what Edge and geckodriver present because
        going into that direction is safer due to the use of more
        precise names. Going the other way could cause name clashes in
        the future.

        Note that this class is meant to be used to normalize
        capabilities **read** from a ``WebDriver`` object. It is not
        meant to be used to normalize capabilities passed for creating
        such an object.

        :param caps: The original capabilities from which to create
                     a normalized capabilities dictionary.
        """
        # Keep a copy for debugging purposes.
        self.caps = caps

        newcaps = dict(caps)
        if "platformName" not in newcaps:
            newcaps["platformName"] = newcaps["platform"]
            del newcaps["platform"]
            newcaps["browserVersion"] = newcaps["version"]
            del newcaps["version"]

        super(NormalizedCapabilities, self).__init__(newcaps)



from selenium import webdriver
import collections

class ConfigTuple(collections.namedtuple(
        'ConfigTuple',
        ('platform', 'browser', 'version'))):

    def as_parameter(self, separator=","):
        return self.platform + separator + self.browser + separator \
            + self.version

configs = collections.OrderedDict()
_configs_by_platform = {}
_configs_by_browser = {}
_configs_by_version = {}

_BROWSER_ABBRS = {
    "IE": "INTERNETEXPLORER",
    "FF": "FIREFOX",
    "CH": "CHROME"
}

def get_config(platform=None, browser=None, version=None):
    browser = browser.upper() if browser is not None else None
    platform = platform.upper() if platform is not None else None

    # Resolve abbreviation if it exists...
    browser = _BROWSER_ABBRS.get(browser, browser) \
        if browser is not None else None

    if platform is not None and browser is not None and version is not None:
        return configs[ConfigTuple(platform, browser, version)]

    ret = None
    if browser is not None:
        if browser not in _configs_by_browser:
            raise ValueError("no configuration for browser: " + browser)
        ret = _configs_by_browser[browser]

    if version is not None:
        if version not in _configs_by_version:
            raise ValueError("no configuration for version: " + version)
        by_version = _configs_by_version[version]
        ret = by_version if ret is None else ret & by_version

    if platform is not None:
        if platform not in _configs_by_platform:
            raise ValueError("no configuration for platform: " + platform)
        by_platform = _configs_by_platform[platform]
        ret = by_platform if ret is None else ret & by_platform

    if len(ret) == 0:
        raise ValueError("no configuration for the combination: {0}, {1}, {2}"
                         .format(platform, browser, version))
    elif len(ret) > 1:
        raise ValueError("the combination {0}, {1}, {2} is ambiguous"
                         .format(platform, browser, version))

    return ret.pop()


def forget():
    # pylint: disable=global-statement
    global configs, _configs_by_platform, _configs_by_browser, \
        _configs_by_version
    configs = {}
    _configs_by_platform = {}
    _configs_by_browser = {}
    _configs_by_version = {}


class Config(object):

    def __init__(self, platform, browser, version, desired_capabilities=None,
                 remote=False):

        if desired_capabilities is None:
            desired_capabilities = {}

        browser = browser.upper() if browser is not None else None
        platform = platform.upper() if platform is not None else None

        # Resolve abbreviation if it exists...
        browser = _BROWSER_ABBRS.get(browser, browser)

        self.platform = platform
        self.browser = browser
        self.version = version
        self.remote = remote
        self.desired_capabilities = desired_capabilities

        key = ConfigTuple(platform, browser, version)
        old = configs.get(key, None)
        configs[key] = self

        ps = _configs_by_platform.setdefault(platform, set())
        ps.add(self)

        bs = _configs_by_browser.setdefault(browser, set())
        bs.add(self)

        vs = _configs_by_version.setdefault(version, set())
        vs.add(self)

        if old:
            ps.remove(old)
            bs.remove(old)
            vs.remove(old)

    def make_selenium_desired_capabilities(self):
        ret = dict(getattr(webdriver.DesiredCapabilities, self.browser))

        ret.update(self.desired_capabilities)
        ret["platform"] = self.platform
        ret["version"] = self.version
        return ret

    def __unicode__(self):
        return "Selenic configured for " + \
            ", ".join((self.platform, self.browser, self.version,
                       "Remote" if self.remote else "Local"))

    def __str__(self):
        return unicode(self)



import subprocess
import os
import tempfile
import time
import httplib
import base64
import shutil
import signal
try:
    import json
except ImportError:
    import simplejson as json

from .base import Remote

def set_test_status(jobid, credentials, passed=True):
    """
    Sets the final status for test runs that are executed at
    SauceLabs.

    :param jobid: The job id for which to set the status.
    :type jobid: :class:`basestring`
    :param credentials: The "user:key" string.
    :type credentials: :class:`basestring`
    :param passed: Whether or not the test run passed. Defaults to ``True``.
    :type passed: :class:`bool`
    :raises Exception: When it can't set the status.
    """
    (username, key) = credentials.split(":")
    creds = base64.encodestring('%s:%s' % (username, key))[:-1]

    conn = httplib.HTTPConnection("saucelabs.com")
    conn.request('PUT', '/rest/v1/%s/jobs/%s' % (username, jobid),
                 json.dumps({"passed": passed}),
                 headers={"Authorization": "Basic " + creds})
    resp = conn.getresponse()
    if resp.status != 200:
        raise Exception("got response: " + resp.status)

def get_unused_sauce_port():
    """
    This returns an unused port among those that Sauce Connect
    forwards. Note that this function does **not** lock the port!

    :returns: A presumably free port that Sauce Connect forwards.
    :rtype: int
    """
    #
    # We exclude 80, 443, 888 from the list since they are reserved.
    #
    sc_ports = [
        8000, 8001, 8003, 8031, 8080, 8081, 8765, 8777,
        8888, 9000, 9001, 9080, 9090, 9876, 9877, 9999,
        49221, 55001]

    for candidate in sc_ports:
        candidate_str = str(candidate)

        # lsof will report a hit whenever *either* the source *or*
        # the destination is using the port we are interested
        # in. There does not seem to be a reliable way to tell
        # lsof to just get ports used on the source side
        # (localhost). We'd have to get all addresses on which the
        # host can listen and list them all. Rather than do this,
        # we check the output to see whether the port is used
        # locally.
        try:
            out = subprocess.check_output(["lsof", "-n", "-P", "-i",
                                           ":" + candidate_str])
        except subprocess.CalledProcessError as ex:
            # When lsof returns an empty list, the exit code is 1,
            # even if there was no actual error. We handle this
            # here.
            if ex.returncode != 1:
                # returncode other than 1 is a real error...
                raise
            # else we swallow the exception...
            out = ""

        used = False
        # Slice in the next line to skip the header line.
        for line in out.splitlines(True)[1:]:
            # Grab the NAME column.
            name = line.split()[8]
            # The file is of the form ``source->destination``. We
            # only care about the source.
            src = name.split("->")[0]

            if src.endswith(":" + candidate_str):
                # We've found that we are actually using the port...
                used = True
                break

        if not used:
            port = candidate
            break

    return port

class Tunnel(object):
    def __init__(self, path, user, key):
        self.path = path
        self.user = user
        self.key = key
        self.process = None
        self.tmpdir = None
        self.tunnel_id = "sc-tunnel-for-" + str(os.getpid())

    def start(self):
        self.tmpdir = tmpdir = tempfile.mkdtemp()
        pidfile_path = os.path.join(tmpdir, "pid")
        logfile_path = os.path.join(tmpdir, "log")
        readyfile_path = os.path.join(tmpdir, "ready")

        tunnel = subprocess.Popen(
            [self.path, "-u", self.user, "-k", self.key,
             "--se-port", "0", "--logfile", logfile_path,
             "--pidfile", pidfile_path, "--readyfile",
             readyfile_path, "--tunnel-identifier", self.tunnel_id])
        while True:
            if os.path.exists(readyfile_path):
                break
            if tunnel.poll():
                raise Exception("tunnel exited prematurely")
            time.sleep(0.2)

        self.process = tunnel

        return self.tunnel_id

    def stop(self):
        if self.process:
            self.process.send_signal(signal.SIGTERM)
            self.process = None
            shutil.rmtree(self.tmpdir, True)

class SauceLabs(Remote):
    name = "saucelabs"
    url_template = "http://{credentials}@ondemand.saucelabs.com:80/wd/hub"

    def __init__(self, *args, **kwargs):
        super(SauceLabs, self).__init__(*args, **kwargs)
        self.tunnel = None

    def build_driver(self, capabilities):
        caps = capabilities
        if self.tunnel_id and self.tunnel:
            raise Exception("you have set a tunnel id and have "
                            "requested a tunnel; you should do one or the "
                            "other")

        if self.tunnel_id or self.tunnel:
            caps['tunnel-identifier'] = self.tunnel_id or \
                                        self.tunnel.tunnel_id

        return super(SauceLabs, self).build_driver(capabilities)

    @property
    def credentials(self):
        return self.conf["SAUCELABS_CREDENTIALS"]

    def set_test_status(self, passed=True):
        """
        Sets the final status for test runs that are executed at
        SauceLabs.

        :param passed: Whether or not the test run passed.
        :type passed: :class:`bool`
        :raises Exception: When it can't set the status.
        """
        set_test_status(self.driver.session_id, self.credentials, passed)

    def get_unused_port(self):
        return get_unused_sauce_port()

    def start_tunnel(self):
        user, key = self.credentials.split(":")
        self.tunnel = Tunnel(self.conf["SC_TUNNEL_PATH"], user, key)
        return self.tunnel.start()

    def stop_tunnel(self):
        self.tunnel.stop()




from .saucelabs import SauceLabs
from .browserstack import BrowserStack

NAME_TO_CLASS={cls.name: cls for cls in (SauceLabs, BrowserStack)}

def get_service_cls(name):
    try:
        return NAME_TO_CLASS[name]
    except IndexError:
        raise ValueError("the service named '{0}' is unknown".format(name))



import socket

def get_unused_port():
    """
    This returns an unused port from the OS. Note that there is no
    guarantee after the port is returned that the OS won't immediately
    reuse it for something else. This function does **not** lock the
    port! On a Linux system which is not starved for ports, there's a
    good chance that the port will still be free by the time the
    calling code uses it.

    The function here is *good enough* for test suites that use
    Selenic. Don't use this for production code.

    :returns: A free port.
    :rtype: int
    """
    # Obtain an unused port from the OS. Linux will assign a port and
    # won't reuse it for a while after we close it.
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.bind(('', 0))
    port = sock.getsockname()[1]
    sock.close()
    return port
